# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
from contextlib import contextmanager
from unittest import mock

import pytest
from immutabledict import immutabledict

from mozphab import exceptions, mozphab, repository, simplecache
from mozphab.commits import Commit
from mozphab.conduit import ConduitAPIError, conduit
from mozphab.diff import Diff
from tests.conftest import search_rev


class Repo:
    api_url = "https://api_url"
    dot_path = "dot_path"
    phab_url = "phab_url"
    path = "path"
    cvs = "git"


def test_set_args_from_repo():
    repo = Repo()
    mozphab.conduit.set_repo(repo)
    assert mozphab.conduit.repo == repo


@pytest.mark.no_mock_token
@mock.patch("mozphab.conduit.read_json_field")
def test_load_api_token(m_read):
    m_read.return_value = False
    mozphab.conduit.set_repo(Repo())
    with pytest.raises(ConduitAPIError):
        mozphab.conduit.load_api_token()

    m_read.return_value = "x"
    assert mozphab.conduit.load_api_token() == "x"


@pytest.mark.no_mock_token
def test_load_api_token_from_env(monkeypatch):
    monkeypatch.setenv("MOZPHAB_PHABRICATOR_API_TOKEN", "api-token123")

    mozphab.conduit.set_repo(Repo())
    token = mozphab.conduit.load_api_token()

    assert token == "api-token123", "API token should be loaded from environment."


@mock.patch("mozphab.conduit.ConduitAPI.load_api_token")
def test_build_request(m_load_api_token):
    m_load_api_token.return_value = "saved-token"
    mozphab.conduit.set_repo(Repo())

    # default token
    assert mozphab.conduit._build_request(
        method="method",
        args={"call": "args"},
        token=None,
    ) == {
        "data": (
            b"params=%7B%22"
            b"call%22%3A%22args%22%2C%22"
            b"__conduit__%22%3A%7B%22"
            b"token%22%3A%22saved-token%22%7D%7D&"
            b"output=json&"
            b"__conduit__=True"
        ),
        "method": "POST",
        "url": "https://api_url/method",
        "headers": {"User-Agent": mock.ANY},
    }

    # provided token
    assert mozphab.conduit._build_request(
        method="method",
        args={"call": "args"},
        token="my-token",
    ) == {
        "data": (
            b"params=%7B%22"
            b"call%22%3A%22args%22%2C%22"
            b"__conduit__%22%3A%7B%22"
            b"token%22%3A%22my-token%22%7D%7D&"
            b"output=json&"
            b"__conduit__=True"
        ),
        "method": "POST",
        "url": "https://api_url/method",
        "headers": {"User-Agent": mock.ANY},
    }

    # unicode
    assert mozphab.conduit._build_request(
        method="method",
        args={"call": "ćwikła"},
        token=None,
    ) == {
        "data": (
            b"params=%7B%22"
            b"call%22%3A%22%5Cu0107wik%5Cu0142a%22%2C%22"
            b"__conduit__%22%3A%7B%22"
            b"token%22%3A%22saved-token%22%7D%7D&"
            b"output=json&"
            b"__conduit__=True"
        ),
        "method": "POST",
        "url": "https://api_url/method",
        "headers": {"User-Agent": mock.ANY},
    }

    # empty dict, empty list
    assert mozphab.conduit._build_request(
        method="method",
        args={"empty_dict": {}, "empty_list": []},
        token=None,
    ) == {
        "data": (
            b"params=%7B%22"
            b"empty_dict%22%3A%7B%7D%2C%22empty_list%22%3A%5B%5D%2C%22"
            b"__conduit__%22%3A%7B%22"
            b"token%22%3A%22saved-token%22%7D%7D&"
            b"output=json&"
            b"__conduit__=True"
        ),
        "method": "POST",
        "url": "https://api_url/method",
        "headers": {"User-Agent": mock.ANY},
    }


@mock.patch("urllib.request.urlopen")
@mock.patch("mozphab.conduit.ConduitAPI.load_api_token")
def test_call(m_load_api_token, m_urlopen):
    m_load_api_token.return_value = "token"
    mozphab.conduit.set_repo(Repo())

    # build fake context-manager to mock urlopen
    cm = mock.MagicMock()
    cm.getcode.return_value = 200
    cm.__enter__.return_value = cm
    m_urlopen.return_value = cm

    # success
    cm.read.return_value = json.dumps({"result": "result", "error_code": False})
    assert mozphab.conduit.call("method", {"call": "args"}) == "result"

    # error
    cm.read.return_value = json.dumps({"error_info": "aieee", "error_code": 1})
    with pytest.raises(ConduitAPIError) as conduit_error:
        mozphab.conduit.call("method", {"call": "args"})
    assert conduit_error.value.args[0].startswith("Phabricator Error: ")


@mock.patch("mozphab.conduit.ConduitAPI.call")
def test_ping(m_call):
    m_call.return_value = {}
    assert mozphab.conduit.ping()

    m_call.side_effect = ConduitAPIError
    assert not mozphab.conduit.ping()

    m_call.side_effect = exceptions.CommandError
    assert not mozphab.conduit.ping()


@mock.patch("mozphab.conduit.ConduitAPI.call")
@mock.patch("mozphab.conduit.ConduitAPI.ping")
@mock.patch("mozphab.conduit.os")
@mock.patch("builtins.open")
def test_check(m_open, m_os, m_ping, m_call):
    check = mozphab.conduit.check

    m_os.path.join.return_value = "x"
    m_os.path.isfile.return_value = True
    assert check()
    m_os.utimie.assert_not_called()

    m_os.path.isfile.return_value = False
    m_ping.return_value = True
    assert check()
    m_open.assert_called_once_with("x", "a")
    m_os.utime.assert_called_once_with("x", None)

    m_ping.return_value = False
    assert not check()


@pytest.fixture
def get_revs():
    mozphab.conduit.set_repo(repository.Repository("", "", "dummy"))
    return mozphab.conduit.get_revisions


@pytest.fixture
def m_call(request):
    request.addfinalizer(simplecache.cache.reset)
    with mock.patch("mozphab.conduit.ConduitAPI.call") as xmock:
        yield xmock


def test_get_revisions_both_ids_and_phids_fails(get_revs, m_call):
    with pytest.raises(ValueError):
        get_revs(ids=[1], phids=["PHID-1"])


def test_get_revisions_none_ids_fails(get_revs, m_call):
    with pytest.raises(ValueError):
        get_revs(ids=None)


def test_get_revisions_none_phids_fails(get_revs, m_call):
    with pytest.raises(ValueError):
        get_revs(phids=None)


basic_phab_result = immutabledict({"data": [{"id": 1, "phid": "PHID-1"}]})


def test_get_revisions_search_by_revid(get_revs, m_call):
    """differential.revision.search by revision-id"""
    m_call.return_value = basic_phab_result

    assert len(get_revs(ids=[1])) == 1
    m_call.assert_called_with(
        "differential.revision.search",
        {"constraints": {"ids": [1]}, "attachments": {"reviewers": True}},
    )


def test_get_revisions_search_by_phid(get_revs, m_call):
    """differential.revision.search by phid"""
    m_call.return_value = basic_phab_result

    assert len(get_revs(phids=["PHID-1"])) == 1
    m_call.assert_called_with(
        "differential.revision.search",
        {"constraints": {"phids": ["PHID-1"]}, "attachments": {"reviewers": True}},
    )


def test_get_revisions_search_by_revid_with_dups(get_revs, m_call):
    """differential.revision.search by revision-id with duplicates"""
    m_call.return_value = basic_phab_result

    assert len(get_revs(ids=[1, 1])) == 2
    m_call.assert_called_with(
        "differential.revision.search",
        {"constraints": {"ids": [1]}, "attachments": {"reviewers": True}},
    )


def test_get_revisions_search_by_phid_with_dups(get_revs, m_call):
    """differential.revision.search by phid with duplicates"""
    m_call.return_value = basic_phab_result

    assert len(get_revs(phids=["PHID-1", "PHID-1"])) == 2
    m_call.assert_called_with(
        "differential.revision.search",
        {"constraints": {"phids": ["PHID-1"]}, "attachments": {"reviewers": True}},
    )


multiple_phab_result = immutabledict(
    {
        "data": [
            {"id": 1, "phid": "PHID-1"},
            {"id": 2, "phid": "PHID-2"},
            {"id": 3, "phid": "PHID-3"},
        ]
    }
)


def test_get_revisions_search_by_revids_ordering(get_revs, m_call):
    """ordering of results must match input when querying by revids"""
    m_call.return_value = multiple_phab_result
    assert get_revs(ids=[2, 1, 3]) == [
        {"id": 2, "phid": "PHID-2"},
        {"id": 1, "phid": "PHID-1"},
        {"id": 3, "phid": "PHID-3"},
    ]


def test_get_revisions_search_by_phids_ordering(get_revs, m_call):
    """ordering of results must match input when querying by phids"""
    m_call.return_value = multiple_phab_result
    assert get_revs(phids=["PHID-2", "PHID-1", "PHID-3"]) == [
        {"id": 2, "phid": "PHID-2"},
        {"id": 1, "phid": "PHID-1"},
        {"id": 3, "phid": "PHID-3"},
    ]


def test_get_revisions_search_by_revids_missing(get_revs, m_call):
    """phabricator does not return info on all rev ids"""
    m_call.return_value = multiple_phab_result
    assert get_revs(ids=[2, 4, 1, 3]) == [
        {"id": 2, "phid": "PHID-2"},
        {"id": 1, "phid": "PHID-1"},
        {"id": 3, "phid": "PHID-3"},
    ]


@pytest.fixture
def get_diffs():
    mozphab.conduit.set_repo(repository.Repository("", "", "dummy"))
    return mozphab.conduit.get_diffs


def test_get_diffs_both_ids_and_phids_fails(get_diffs, m_call):
    with pytest.raises(ValueError):
        get_diffs(ids=[1], phids=["PHID-1"])


def test_get_diffs_none_ids_fails(get_diffs, m_call):
    with pytest.raises(ValueError):
        get_diffs(ids=None)


def test_get_diffs_none_phids_fails(get_diffs, m_call):
    with pytest.raises(ValueError):
        get_diffs(phids=None)


def test_get_diffs_by_phid(get_diffs, m_call):
    m_call.return_value = basic_phab_result

    assert (
        len(get_diffs(phids=["PHID-1"])) == 1
    ), "Should be a length of 1 when given 1 PHID"
    m_call.assert_called_with(
        "differential.diff.search",
        {"constraints": {"phids": ["PHID-1"]}, "attachments": {"commits": True}},
    )


def test_get_diffs_by_id(get_diffs, m_call):
    m_call.return_value = basic_phab_result

    assert len(get_diffs(ids=[1])) == 1, "Should be a length of 1 when given 1 ID"
    m_call.assert_called_with(
        "differential.diff.search",
        {"constraints": {"ids": [1]}, "attachments": {"commits": True}},
    )


def test_get_diffs_search_by_phid_with_dups(get_diffs, m_call):
    """differential.diff.search by phid with duplicates"""
    m_call.return_value = basic_phab_result

    assert (
        len(get_diffs(phids=["PHID-1", "PHID-1"])) == 1
    ), "Should not include redundant/duplicate diffs via PHID"
    m_call.assert_called_with(
        "differential.diff.search",
        {"constraints": {"phids": ["PHID-1"]}, "attachments": {"commits": True}},
    )


def test_get_diffs_search_by_diffid_with_dups(get_diffs, m_call):
    """differential.diff.search by diff-id with duplicates"""
    m_call.return_value = basic_phab_result

    assert (
        len(get_diffs(ids=[1, 1])) == 1
    ), "Should not include redundant/duplicate diffs via ID"
    m_call.assert_called_with(
        "differential.diff.search",
        {"constraints": {"ids": [1]}, "attachments": {"commits": True}},
    )


def test_get_diffs_search_by_diffids_missing(get_diffs, m_call):
    """phabricator does not return info on all diff ids"""
    m_call.return_value = multiple_phab_result

    diff_dict = get_diffs(ids=[2, 4, 1, 3])
    assert diff_dict.get("PHID-1") == {
        "id": 1,
        "phid": "PHID-1",
    }, "Should return dict of Diff 1"
    assert diff_dict.get("PHID-2") == {
        "id": 2,
        "phid": "PHID-2",
    }, "Should return dict of Diff 2"
    assert diff_dict.get("PHID-3") == {
        "id": 3,
        "phid": "PHID-3",
    }, "Should return dict of Diff 3"
    assert not diff_dict.get("PHID-4"), "Should not return dict of non-existent diff"


@mock.patch("mozphab.conduit.ConduitAPI.call")
def test_get_related_phids(m_call):
    get_related_phids = mozphab.conduit.get_related_phids

    m_call.return_value = {}
    assert [] == get_related_phids("aaa", include_abandoned=True)
    m_call.assert_called_once_with(
        "edge.search", {"sourcePHIDs": ["aaa"], "types": ["revision.parent"]}
    )

    m_call.side_effect = [
        {"data": [{"destinationPHID": "bbb"}]},
        {"data": [{"destinationPHID": "aaa"}]},
        {},
    ]
    assert ["bbb", "aaa"] == get_related_phids("ccc", include_abandoned=True)

    m_call.side_effect = [
        {"data": [{"destinationPHID": "bbb"}]},
        {"data": [{"destinationPHID": "aaa"}]},
        {},
        {
            "data": [
                {"id": 1, "phid": "aaa", "fields": {"status": {"value": "-"}}},
                {"id": 2, "phid": "bbb", "fields": {"status": {"value": "abandoned"}}},
            ]
        },
    ]
    assert ["aaa"] == get_related_phids("ccc", include_abandoned=False)


def test_has_revision_reviewers(m_call):
    commit = Commit(rev_id=None)
    assert not conduit.has_revision_reviewers(commit)

    m_call.return_value = {"data": [search_rev(rev=123, reviewers=[])]}
    commit = Commit(rev_id=123)
    assert not conduit.has_revision_reviewers(commit)

    m_call.return_value = {"data": [search_rev(rev=456, reviewers=["alice"])]}
    commit = Commit(rev_id=456)
    assert conduit.has_revision_reviewers(commit)


@mock.patch("builtins.open")
@mock.patch("mozphab.conduit.json")
@mock.patch("mozphab.conduit.get_arcrc_path")
@mock.patch("os.chmod")
def test_save_api_token(m_chmod, m_get_arcrc_path, m_json, m_open, git):
    save_api_token = conduit.save_api_token

    @contextmanager
    def with_open():
        yield None

    m_get_arcrc_path.return_value = ".arcrc"
    git.api_url = "http://test/api/"
    mozphab.conduit.set_repo(git)
    m_open.side_effect = PermissionError
    with pytest.raises(PermissionError):
        save_api_token("abc")

    m_chmod.reset_mock()
    m_open.side_effect = (FileNotFoundError, with_open())
    save_api_token("abc")
    m_chmod.assert_called_once_with(".arcrc", 0o600)

    m_json.dump.assert_called_once_with(
        {"hosts": {git.api_url: {"token": "abc"}}}, mock.ANY, sort_keys=True, indent=2
    )

    m_chmod.reset_mock()
    m_json.reset_mock()
    m_open.side_effect = None
    m_json.load.return_value = {"existing_key": "existing_value"}
    save_api_token("abc")
    m_json.dump.assert_called_once_with(
        {"hosts": {git.api_url: {"token": "abc"}}, "existing_key": "existing_value"},
        mock.ANY,
        sort_keys=True,
        indent=2,
    )
    m_chmod.assert_not_called()

    m_json.reset_mock()
    m_json.load.return_value = {
        "hosts": {git.api_url: {"token": "token1"}, "address2": {"token": "token2"}},
        "existing_key": "existing_value",
    }
    save_api_token("abc")
    m_json.dump.assert_called_once_with(
        {
            "hosts": {git.api_url: {"token": "abc"}, "address2": {"token": "token2"}},
            "existing_key": "existing_value",
        },
        mock.ANY,
        sort_keys=True,
        indent=2,
    )


def test_parse_git_diff():
    parse = Diff.parse_git_diff
    assert parse("@@ -40,9 +50,3 @@ packaging==19.1 \\") == (40, 50, 9, 3)


@mock.patch("mozphab.repository.conduit.call")
def test_diff_property(m_call, git, hg):
    git.get_public_node = lambda x: x
    git._phab_vcs = "git"
    conduit.set_repo(git)
    commit = Commit(
        name="abc-name",
        author_name="Author Name",
        author_email="auth@or.email",
        author_date_epoch=1234567,
        title_preview="Title Preview",
        node="abc",
        parent="def",
        wip=False,
    )
    mozphab.conduit.set_diff_property("1", commit, "message")
    m_call.assert_called_once_with(
        "differential.setdiffproperty",
        {
            "diff_id": "1",
            "name": "local:commits",
            "data": json.dumps(
                {
                    "abc": {
                        "author": "Author Name",
                        "authorEmail": "auth@or.email",
                        "time": 1234567,
                        "summary": "Title Preview",
                        "message": "message",
                        "commit": "abc",
                        "parents": ["def"],
                    }
                }
            ),
        },
    )

    m_call.reset_mock()
    git._phab_vcs = "hg"
    git._cinnabar_installed = True
    mozphab.conduit.set_diff_property("1", commit, "message")
    m_call.assert_called_once_with(
        "differential.setdiffproperty",
        {
            "diff_id": "1",
            "name": "local:commits",
            "data": json.dumps(
                {
                    "abc": {
                        "author": "Author Name",
                        "authorEmail": "auth@or.email",
                        "time": 1234567,
                        "summary": "Title Preview",
                        "message": "message",
                        "commit": "abc",
                        "parents": ["def"],
                        "rev": "abc",
                    }
                }
            ),
        },
    )

    m_call.reset_mock()
    hg._phab_vcs = "hg"
    mozphab.conduit.set_repo(hg)
    mozphab.conduit.set_diff_property("1", commit, "message")
    m_call.assert_called_once_with(
        "differential.setdiffproperty",
        {
            "diff_id": "1",
            "name": "local:commits",
            "data": json.dumps(
                {
                    "abc": {
                        "author": "Author Name",
                        "authorEmail": "auth@or.email",
                        "time": 1234567,
                        "summary": "Title Preview",
                        "message": "message",
                        "commit": "abc",
                        "parents": ["def"],
                        "rev": "abc",
                    }
                }
            ),
        },
    )


@mock.patch("mozphab.repository.conduit.call")
def test_get_projects(m_call):
    expected_projects = [
        {
            "id": 1,
            "type": "PROJ",
            "phid": "PHID-PROJ-1",
            "attachments": {},
            "fields": {"name": "A", "slug": "a"},
        },
        {
            "id": 2,
            "type": "PROJ",
            "phid": "PHID-PROJ-2",
            "attachments": {},
            "fields": {"name": "B", "slug": "b"},
        },
    ]
    m_call.side_effect = (
        {
            "data": expected_projects,
            "maps": {},
            "query": {"queryKey": None},
            "cursor": {"limit": 100, "after": None, "before": None, "order": None},
        },
    )
    projects = mozphab.conduit.get_projects(["a", "b"])
    m_call.assert_called_once_with(
        "project.search", {"constraints": {"slugs": ["a", "b"]}}
    )
    assert projects == expected_projects


@mock.patch("mozphab.repository.conduit.get_projects")
def test_get_project_phid(m_get_projects):
    m_get_projects.side_effect = (
        [
            {
                "id": 99,
                "type": "PROJ",
                "phid": "PHID-PROJ-1",
                "fields": {
                    "name": "Release managers",
                    "slug": "release-managers",
                    "subtype": "default",
                    "milestone": None,
                    "depth": 0,
                    "parent": None,
                    "icon": {"key": "tag", "name": "Tag", "icon": "fa-tags"},
                    "color": {"key": "blue", "name": "Blue"},
                    "spacePHID": None,
                    "dateCreated": 1552593979,
                    "dateModified": 1558110482,
                    "policy": {"view": "public", "edit": "admin", "join": "no-one"},
                    "description": "description",
                },
                "attachments": {},
            }
        ],
    )
    phid = mozphab.conduit.get_project_phid("release-managers")
    m_get_projects.assert_called_once_with(["release-managers"])
    assert phid == "PHID-PROJ-1"


@mock.patch("mozphab.repository.conduit.call")
def test_submit_diff(m_call, git):
    mozphab.conduit.set_repo(git)
    diff = Diff()
    commit = Commit(
        name="abc-name",
        author_name="Author Name",
        author_email="auth@example.com",
        author_date_epoch=1234567,
        title_preview="Title Preview",
        body="Additional summary.",
        bug_id="777",
        rev_id=456,
        node="abc",
        parent="def",
    )
    m_call.side_effect = [
        # differential.revision.search
        {"data": [search_rev(rev=456, repo="PHID-BETA")]},
        # differential.creatediff
        {},
    ]
    mozphab.conduit.submit_diff(diff, commit)
    m_call.assert_any_call(
        "differential.creatediff",
        {
            "changes": mock.ANY,
            "sourceMachine": mock.ANY,
            "sourceControlSystem": "git",
            "sourceControlPath": "/",
            "sourceControlBaseRevision": "def",
            "creationMethod": "moz-phab-git",
            "lintStatus": "none",
            "unitStatus": "none",
            "repositoryPHID": "PHID-BETA",
            "sourcePath": mock.ANY,
            "branch": "HEAD",
        },
    )


@mock.patch("mozphab.repository.conduit.call")
def test_submit_diff_uplift(m_call, git):
    """Test that uplift command adds 'uplift' to creationMethod."""
    git.args.command = "uplift"
    mozphab.conduit.set_repo(git)
    diff = Diff()
    commit = Commit(
        name="abc-name",
        author_name="Author Name",
        author_email="auth@example.com",
        author_date_epoch=1234567,
        title_preview="Title Preview",
        body="Additional summary.",
        bug_id="777",
        rev_id=456,
        node="abc",
        parent="def",
    )
    m_call.side_effect = [
        # differential.revision.search
        {"data": [search_rev(rev=456, repo="PHID-BETA")]},
        # differential.creatediff
        {},
    ]
    mozphab.conduit.submit_diff(diff, commit)
    m_call.assert_any_call(
        "differential.creatediff",
        {
            "changes": mock.ANY,
            "sourceMachine": mock.ANY,
            "sourceControlSystem": "git",
            "sourceControlPath": "/",
            "sourceControlBaseRevision": "def",
            "creationMethod": "moz-phab-git-uplift",
            "lintStatus": "none",
            "unitStatus": "none",
            "repositoryPHID": "PHID-BETA",
            "sourcePath": mock.ANY,
            "branch": "HEAD",
        },
    )


@mock.patch("mozphab.repository.conduit.call")
def test_submit_diff_lando(m_call, git, monkeypatch):
    """Test that MOZPHAB_PHABRICATOR_API_TOKEN env var adds 'lando' to creationMethod."""
    monkeypatch.setenv("MOZPHAB_PHABRICATOR_API_TOKEN", "some-token")
    mozphab.conduit.set_repo(git)
    diff = Diff()
    commit = Commit(
        name="abc-name",
        author_name="Author Name",
        author_email="auth@example.com",
        author_date_epoch=1234567,
        title_preview="Title Preview",
        body="Additional summary.",
        bug_id="777",
        rev_id=456,
        node="abc",
        parent="def",
    )
    m_call.side_effect = [
        # differential.revision.search
        {"data": [search_rev(rev=456, repo="PHID-BETA")]},
        # differential.creatediff
        {},
    ]
    mozphab.conduit.submit_diff(diff, commit)
    m_call.assert_any_call(
        "differential.creatediff",
        {
            "changes": mock.ANY,
            "sourceMachine": mock.ANY,
            "sourceControlSystem": "git",
            "sourceControlPath": "/",
            "sourceControlBaseRevision": "def",
            "creationMethod": "moz-phab-git-lando",
            "lintStatus": "none",
            "unitStatus": "none",
            "repositoryPHID": "PHID-BETA",
            "sourcePath": mock.ANY,
            "branch": "HEAD",
        },
    )


@mock.patch("mozphab.repository.conduit.call")
def test_submit_diff_uplift_and_lando(m_call, git, monkeypatch):
    """Test that both uplift command and lando env var are reflected in creationMethod."""
    git.args.command = "uplift"
    monkeypatch.setenv("MOZPHAB_PHABRICATOR_API_TOKEN", "some-token")
    mozphab.conduit.set_repo(git)
    diff = Diff()
    commit = Commit(
        name="abc-name",
        author_name="Author Name",
        author_email="auth@example.com",
        author_date_epoch=1234567,
        title_preview="Title Preview",
        body="Additional summary.",
        bug_id="777",
        rev_id=456,
        node="abc",
        parent="def",
    )
    m_call.side_effect = [
        # differential.revision.search
        {"data": [search_rev(rev=456, repo="PHID-BETA")]},
        # differential.creatediff
        {},
    ]
    mozphab.conduit.submit_diff(diff, commit)
    m_call.assert_any_call(
        "differential.creatediff",
        {
            "changes": mock.ANY,
            "sourceMachine": mock.ANY,
            "sourceControlSystem": "git",
            "sourceControlPath": "/",
            "sourceControlBaseRevision": "def",
            "creationMethod": "moz-phab-git-uplift-lando",
            "lintStatus": "none",
            "unitStatus": "none",
            "repositoryPHID": "PHID-BETA",
            "sourcePath": mock.ANY,
            "branch": "HEAD",
        },
    )


@mock.patch("mozphab.git.Git.get_public_node")
@mock.patch("mozphab.repository.conduit.call")
def test_submit_diff_cinnabar_uplift_lando(m_call, m_get_public_node, git, monkeypatch):
    """Test that cinnabar, uplift, and lando all appear in creationMethod in correct order."""
    git.args.command = "uplift"
    git._cinnabar_installed = True
    # Make cinnabar required by setting phab_vcs different from local vcs
    git._phab_vcs = "hg"  # git.vcs is "git", so is_cinnabar_required will be True
    monkeypatch.setenv("MOZPHAB_PHABRICATOR_API_TOKEN", "some-token")
    # Mock get_public_node to just return the node as-is
    m_get_public_node.side_effect = lambda x: x
    mozphab.conduit.set_repo(git)
    diff = Diff()
    commit = Commit(
        name="abc-name",
        author_name="Author Name",
        author_email="auth@example.com",
        author_date_epoch=1234567,
        title_preview="Title Preview",
        body="Additional summary.",
        bug_id="777",
        rev_id=456,
        node="abc",
        parent="def",
    )
    m_call.side_effect = [
        # differential.revision.search
        {"data": [search_rev(rev=456, repo="PHID-BETA")]},
        # differential.creatediff
        {},
    ]
    mozphab.conduit.submit_diff(diff, commit)
    m_call.assert_any_call(
        "differential.creatediff",
        {
            "changes": mock.ANY,
            "sourceMachine": mock.ANY,
            "sourceControlSystem": "hg",  # When phab_vcs is "hg", sourceControlSystem is "hg"
            "sourceControlPath": "/",
            "sourceControlBaseRevision": "def",
            "creationMethod": "moz-phab-git-cinnabar-uplift-lando",
            "lintStatus": "none",
            "unitStatus": "none",
            "repositoryPHID": "PHID-BETA",
            "sourcePath": mock.ANY,
            "branch": "default",  # When phab_vcs is "hg", branch is "default"
        },
    )


@mock.patch("mozphab.repository.conduit.call")
def test_create_revision(m_call):
    commit = Commit(
        name="abc-name",
        author_name="Author Name",
        author_email="auth@example.com",
        author_date_epoch=1234567,
        title_preview="Title Preview",
        body="Additional summary.",
        bug_id="777",
        node="abc",
        parent="def",
        wip=False,
    )
    mozphab.conduit.create_revision(commit, "PHID-DIFF-7")
    m_call.assert_called_once_with(
        "differential.revision.edit",
        {
            "transactions": [
                {"type": "title", "value": "Title Preview"},
                {"type": "summary", "value": "Additional summary."},
                {"type": "bugzilla.bug-id", "value": "777"},
                {"type": "update", "value": "PHID-DIFF-7"},
            ]
        },
    )


class TestEditRevision:
    @staticmethod
    def _get_revisions(status):
        return [{"fields": {"status": {"value": status}}}]

    @mock.patch("mozphab.repository.conduit.call")
    def test_new_wip(self, m_call):
        conduit.edit_revision(wip=True)
        m_call.assert_called_once_with(
            "differential.revision.edit",
            {
                "transactions": [
                    {"type": "plan-changes", "value": True},
                ],
            },
        )

    @mock.patch("mozphab.repository.conduit.call")
    def test_new_no_wip(self, m_call):
        conduit.edit_revision(wip=False)
        m_call.assert_called_once_with(
            "differential.revision.edit",
            {
                "transactions": [],
            },
        )

    @mock.patch("mozphab.repository.conduit.get_revisions")
    @mock.patch("mozphab.repository.conduit.call")
    def test_update_wip_changes_planned(self, m_call, m_get_revisions):
        # wip + changes-planned -> plan-changes
        m_get_revisions.return_value = self._get_revisions("changes-planned")
        conduit.edit_revision(rev_id=1, wip=True)
        call_args = m_call.call_args_list
        assert len(call_args) == 2
        assert call_args[0] == mock.call(
            "differential.revision.edit",
            {
                "transactions": [],
                "objectIdentifier": 1,
            },
        )
        assert call_args[1] == mock.call(
            "differential.revision.edit",
            {
                "transactions": [
                    {"type": "plan-changes", "value": True},
                ],
                "objectIdentifier": 1,
            },
        )

    @mock.patch("mozphab.repository.conduit.get_revisions")
    @mock.patch("mozphab.repository.conduit.call")
    def test_update_wip_needs_review(self, m_call, m_get_revisions):
        # wip + needs-review -> plan-changes
        m_get_revisions.return_value = self._get_revisions("needs-review")
        conduit.edit_revision(rev_id=1, wip=True)
        m_call.assert_called_once_with(
            "differential.revision.edit",
            {
                "transactions": [
                    {"type": "plan-changes", "value": True},
                ],
                "objectIdentifier": 1,
            },
        )

    @mock.patch("mozphab.repository.conduit.get_revisions")
    @mock.patch("mozphab.repository.conduit.call")
    def test_update_wip_accepted(self, m_call, m_get_revisions):
        # wip + accepted -> plan-changes
        m_get_revisions.return_value = self._get_revisions("accepted")
        conduit.edit_revision(rev_id=1, wip=True)
        m_call.assert_called_once_with(
            "differential.revision.edit",
            {
                "transactions": [
                    {"type": "plan-changes", "value": True},
                ],
                "objectIdentifier": 1,
            },
        )

    @mock.patch("mozphab.repository.conduit.get_revisions")
    @mock.patch("mozphab.repository.conduit.call")
    def test_update_no_wip_changes_planned(self, m_call, m_get_revisions):
        # no-wip + changes-planned -> request-review
        m_get_revisions.return_value = self._get_revisions("changes-planned")
        conduit.edit_revision(rev_id=1, wip=False)
        m_call.assert_called_once_with(
            "differential.revision.edit",
            {
                "transactions": [
                    {"type": "request-review", "value": True},
                ],
                "objectIdentifier": 1,
            },
        )

    @mock.patch("mozphab.repository.conduit.get_revisions")
    @mock.patch("mozphab.repository.conduit.call")
    def test_update_no_wip_needs_review(self, m_call, m_get_revisions):
        # no-wip + needs-review -> no-op
        m_get_revisions.return_value = self._get_revisions("needs-review")
        conduit.edit_revision(rev_id=1, wip=False)
        m_call.assert_called_once_with(
            "differential.revision.edit",
            {
                "transactions": [],
                "objectIdentifier": 1,
            },
        )

    @mock.patch("mozphab.repository.conduit.get_revisions")
    @mock.patch("mozphab.repository.conduit.call")
    def test_update_no_wip_accepted(self, m_call, m_get_revisions):
        # no-wip + accepted -> no-op
        m_get_revisions.return_value = self._get_revisions("accepted")
        conduit.edit_revision(rev_id=1, wip=False)
        m_call.assert_called_once_with(
            "differential.revision.edit",
            {
                "transactions": [],
                "objectIdentifier": 1,
            },
        )

    @mock.patch("mozphab.repository.conduit.get_revisions")
    @mock.patch("mozphab.repository.conduit.call")
    def test_edit_revision_sets_parent(self, m_call, m_get_revisions):
        m_get_revisions.return_value = self._get_revisions("accepted")
        conduit.edit_revision(rev_id=1, wip=False, parent_rev_phid="PHID-DREV-123")
        m_call.assert_called_once_with(
            "differential.revision.edit",
            {
                "transactions": [{"type": "parents.add", "value": ["PHID-DREV-123"]}],
                "objectIdentifier": 1,
            },
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pytest

from mozphab.repository import get_lando_url_for_phabricator, is_mozilla_phabricator


def test_is_mozilla_phabricator():
    assert is_mozilla_phabricator("https://phabricator.services.mozilla.com/")
    assert is_mozilla_phabricator("https://phabricator-dev.allizom.org/")
    assert not is_mozilla_phabricator("https://reviews.llvm.org/")


def test_get_lando_url_for_phabricator():
    assert (
        get_lando_url_for_phabricator("https://phabricator.services.mozilla.com/")
        == "https://lando.moz.tools"
    ), "Prod Phabricator with trailing slash should return prod Lando URL."
    assert (
        get_lando_url_for_phabricator("https://phabricator.services.mozilla.com")
        == "https://lando.moz.tools"
    ), "Prod Phabricator without trailing slash should return prod Lando URL."
    assert (
        get_lando_url_for_phabricator("https://phabricator-dev.allizom.org/")
        == "https://dev.lando.nonprod.webservices.mozgcp.net"
    ), "Dev Phabricator should return dev Lando URL."

    with pytest.raises(KeyError):
        get_lando_url_for_phabricator("https://reviews.llvm.org/")



import argparse
import os
import re
import uuid
from datetime import datetime, timezone
from pathlib import Path
from typing import List, Optional, Tuple, Union

from packaging.version import Version

from mozphab import environment

from .commits import Commit
from .config import config
from .diff import Diff
from .exceptions import CommandError, Error
from .git import Git
from .helpers import (
    is_valid_email,
    short_node,
    temporary_binary_file,
    temporary_file,
)
from .logger import logger
from .repository import Repository
from .spinner import wait_message
from .subprocess_wrapper import check_call, check_output, subprocess


class Jujutsu(Repository):
    MIN_VERSION = Version("0.33.0")

    @classmethod
    def is_repo(cls, path: str) -> bool:
        """Quick check for repository at specified path."""
        return os.path.exists(os.path.join(path, ".jj"))

    # ----
    # Methods expected from callers of the `Repository` interface:
    # ----

    def __init__(self, path: str):
        dot_path = os.path.join(path, ".jj")
        if not os.path.exists(dot_path):
            raise ValueError(f"{path}: not a Jujutsu repository")

        logger.debug("found Jujutsu repo in %s", path)

        version_str, version = self.__check_and_get_version()
        self.vcs_version = version_str
        self.__version = version

        resolved_path = Path(path).resolve(strict=True)
        logger.debug(f"resolved_path: {resolved_path}")

        try:
            self.git_path = Path(
                self.__check_output(["jj", "git", "root"], split=False)
            )
        except Exception:
            raise ValueError(
                f"{path}: failed to run `jj git root`, likely not a Jujutsu repository"
            )
        logger.debug(f"git_path: {self.git_path}")

        try:
            is_colocated = (
                resolved_path == self.git_path.parent and self.git_path.name == ".git"
            )
            logger.debug(f"is_colocated: {is_colocated}")
            bare_path = None if is_colocated else str(self.git_path)
            self.__git_repo = Git(path, bare_path=bare_path)
        except Exception:
            raise ValueError(
                f"internal error: failed to initialize Git repo from {self.git_path}"
            )

        # Populate common fields expected from a `Repository`
        super().__init__(path, dot_path)

        self.vcs = "jj"

        self.revset = None
        self.branch = None

        self.__email = self.__check_output(
            ["jj", "config", "get", "user.email"], split=False
        ).rstrip()

    def __check_and_get_version(self) -> Tuple[str, Version]:
        min_version = Jujutsu.MIN_VERSION

        version_re = re.compile(r"jj (\d+\.\d+\.\d+)(?:-[a-fA-F0-9]{40})?")
        try:
            jj_version_output = self.__check_output(["jj", "version"], split=False)
        except FileNotFoundError as exc:
            if exc.filename == "jj":
                raise Error("`jj` executable was not found.")
            raise exc

        m = version_re.fullmatch(jj_version_output)
        if not m:
            raise Error("Failed to determine Jujutsu version.")
        version = Version(m.group(1))

        if version < min_version:
            raise Error(f"`moz-phab` requires Jujutsu {min_version} or higher.")
        return m.group(0), version

    def check_vcs(self) -> bool:
        if self.args.force_vcs:
            return True

        if self.is_cinnabar_required and not self.is_cinnabar_installed:
            raise Error(
                "Git Cinnabar extension is required to work on this repository."
            )

        return True

    def before_submit(self):
        self.__validate_email()

    def set_args(self, args: argparse.Namespace):
        """Store moz-phab command line args and set the revset."""
        super().set_args(args)

        is_single = hasattr(self.args, "single") and self.args.single

        if hasattr(self.args, "start_rev"):
            start_rev = None
            if self.args.start_rev != environment.DEFAULT_START_REV:
                start_rev = self.args.start_rev
            else:
                start_rev = (
                    self.__get_last_stack_change()
                    if is_single
                    else self.__get_first_stack_change()
                )
            if not start_rev:
                return None

            end_rev = None
            if is_single:
                end_rev = start_rev
            elif (
                hasattr(self.args, "end_rev")
                and self.args.end_rev != environment.DEFAULT_END_REV
            ):
                end_rev = self.args.end_rev
            else:
                end_rev = self.__get_last_stack_change()
            if not end_rev:
                return None

            self.revset = (start_rev, end_rev)

    def commit_stack(self, single: bool = False) -> Optional[List[Commit]]:
        """Collect all the info about commits."""
        logger.debug(f"searching with start and end at {self.revset}")
        if not self.revset:
            # No commits found to submit
            return None

        boundary = "--%s--\n" % uuid.uuid4().hex
        log = self.__cli_log(
            use_reversed=True,
            template="".join(
                [
                    "separate('\n', ",
                    "author.timestamp().utc().format('%+'), ",
                    "author.name(), ",
                    "author.email(), ",
                    "self.parents().map(|c| c.commit_id().short()).join(' '), ",
                    "change_id.short(), ",
                    "commit_id, ",
                    "conflict, ",
                    "description, ",
                    f'"{boundary}"',
                    ")",
                ]
            ),
            revset="({})::({})".format(self.revset[0], self.revset[1]),
            split=False,
            strip=False,
        )[: -len(boundary)]
        changes = log.split(boundary)

        commits = []
        first_commit_id = None
        for log_line in changes:
            if not log_line:
                continue  # TODO: Why would we need this? This seems suspicious.

            (
                author_date,
                author_name,
                author_email,
                parents,
                change_id,
                commit_id,
                is_conflicted,
                desc,
            ) = log_line.split("\n", 7)
            desc = desc.splitlines()

            tree_hash = self.__git_repo._revparse(commit_id + ":./")

            parents = parents.split(" ")
            if first_commit_id != commit_id and len(parents) > 1:
                raise Error(
                    f"Multiple parents found for change {change_id}, unable to continue"
                )
            first_commit_id = first_commit_id or commit_id

            is_conflicted = Jujutsu.__parse_log_bool(
                "`conflict` template segment", is_conflicted
            )
            if is_conflicted:
                raise Error(f"Change {change_id} is conflicted, unable to continue")

            commit_epoch = datetime.fromisoformat(author_date).timestamp()
            commits.append(
                Commit(
                    name=change_id,
                    node=commit_id,
                    orig_node=commit_id,
                    submit=True,
                    title=desc[0],
                    title_preview=desc[0],
                    body="\n".join(desc[1:]).rstrip(),
                    bug_id=None,
                    reviewers={"request": [], "granted": []},
                    rev_id=None,
                    parent=parents[0],
                    tree_hash=tree_hash,
                    author_date=author_date,
                    author_date_epoch=commit_epoch,
                    author_name=author_name,
                    author_email=author_email,
                )
            )

        return commits

    def after_submit(self):
        # `moz-phab submit` operations on the Jujutsu backend don't change the change [sic] that the
        # current workspace is working with—including the Git commands we issue, which should be
        # read-only. So, we don't have to do anything. Yay!
        pass

    def untracked(self) -> List[str]:
        is_working_copy_descriptionless_but_changed = self.__cli_log(
            revset="@",
            template="self.description().len() == 0 && !self.empty()",
            split=False,
        )
        is_working_copy_descriptionless_but_changed = Jujutsu.__parse_log_bool(
            "check for descriptionless-but-changed working copy",
            is_working_copy_descriptionless_but_changed,
        )
        if is_working_copy_descriptionless_but_changed:
            return self.__cli_log(
                revset="@",
                template='self.diff().files().map(|f| f.path()).join("\n")',
            )
        return []

    def get_diff(self, commit: Commit) -> Diff:
        """Create a Diff object with changes."""
        # NOTE: If we don't do this, then we break on a `lesscontext` member missing from `args`.
        self.__git_repo.args = self.args
        return self.__git_repo.get_diff(commit)

    def amend_commit(self, commit: Commit, commits: List[Commit]):
        """Amend the commit with an updated message.

        Changing commit's message changes also its SHA1.
        All the children within the stack and bookmarks are then updated
        to keep the history.

        Args:
            commit: Information about the commit to be amended
            commits: List of commits within the stack
        """
        updated_message = f"{commit.title}\n{commit.body}"

        # # TODO: apply this optimization
        # current_body = self.git_out(
        #     ["show", "-s", "--format=%s%n%b", commit.node], split=False
        # )
        # if current_body == updated_body:
        #     logger.debug("not amending commit %s, unchanged", commit.name)
        #     return

        change_id = commit.name
        # TODO: It would be nice to open an issue to drive amending in bulk, rather than
        # `describe`-ing and rebasing all descendants with each change (potentially); `jj describe`
        # on multiple commits can edit the message for _all_ changes in the stack in one go, if we
        # need it.
        with temporary_file(updated_message) as message_path:
            with open(message_path) as message_file:
                check_call(["jj", "describe", change_id, "--stdin"], stdin=message_file)

    def finalize(self, commits: List[Commit]):
        pass

    def cleanup(self):
        # NOTE: This backend only `jj describe`s changes, and doesn't change them, so we haven't
        # changed any state that we might need to restore.
        pass

    def refresh_commit_stack(self, commits: List[Commit]):
        # TODO: update `commit[n].node` to match current `commit_id`s
        pass

    @property
    def is_cinnabar_installed(self) -> bool:
        """Check if Cinnabar extension is callable."""
        return self.__git_repo.is_cinnabar_installed

    @property
    def is_cinnabar_required(self) -> bool:
        """Check if local VCS is different than the remote one."""
        return self.__git_repo.is_cinnabar_required

    def get_public_node(self, node: str) -> str:
        """Return a Mercurial node if Cinnabar is required."""
        return self.__git_repo.get_public_node(node)

    # TODO: Functionality to make `local_uplift_if_possible` work?

    def is_worktree_clean(self) -> bool:
        """Check if the working tree is clean."""
        check_call(["jj", "git", "export"])
        return True

    def check_node(self, node: str) -> str:
        """Check if the node exists.

        Consults `jj log --revisions $node` first, and:

        1. If multiple commits are found, return an error.
        2. If a single commit is found, return it as a node.
        3. If the CLI errors (which happens when there is no match,
           among other things), then fall back to the Git backend's
           behavior.

        Raises NotFoundError if none of the above yield a single commit.
        """
        try:
            commits = self.__cli_log(template='commit_id ++ "\\n"', revset=node)
            if len(commits) > 1:
                raise Error(
                    f"Multiple commits match revset `{node}`, unable to continue"
                )
            return commits[0]
        except CommandError:
            # Call the git backend as well, as it will try to resolve any
            # cinnabar hashes.
            return self.__git_repo.check_node(node)

    def before_patch(self, node: str, name: str):
        """Prepare repository to apply the patches.

        Args:
            node - SHA1 of the base commit
            name - name of the bookmark to be created
        """

        if node:
            with wait_message("Checking out %s.." % short_node(node)):
                check_call(["jj", "new", "--quiet", node])
            logger.info("Checked out %s", short_node(node))

        if name and not self.args.no_branch and config.create_branch:
            branches = set(
                self.__check_output(
                    ["jj", "bookmark", "list", "--template", 'name ++ "\n"'],
                    strip=False,
                )
            )
            branches = [re.sub("[ *]", "", b) for b in branches]
            branch_name = name
            i = 0
            while branch_name in branches:
                i += 1
                branch_name = "%s_%s" % (name, i)

            check_call(
                ["jj", "bookmark", "create", "--quiet", "--revision=@", branch_name]
            )
            logger.info("Created bookmark %s", branch_name)
            self.__patch_branch_name = branch_name

    def apply_patch(
        self, diff: str, body: str, author: Optional[str], author_date: Optional[int]
    ):
        # NOTE: `before_patch` ensures that we are editing a new, empty commit on the base we want.

        # apply the patch as a binary file to ensure the correct line endings
        # is used.
        # TODO: Use `jj`'s built-in patching facilities when it exists (see
        # <https://github.com/martinvonz/jj/issues/2702>).
        with temporary_binary_file(diff.encode("utf8")) as patch_file:
            # NOTE: We avoid `self.__git_repo.git_call` because it changes the CWD.
            self.__git_repo.git.call(["apply", patch_file], cwd=self.path)

        # TODO: dedupe with other `describe` usage
        with temporary_file(body) as message_path:
            with open(message_path) as message_file:
                check_call(["jj", "describe", "--quiet", "--stdin"], stdin=message_file)
        author_date = datetime.fromtimestamp(author_date, tz=timezone.utc).isoformat()
        check_call(
            [
                "jj",
                "metaedit",
                "--quiet",
                "--author",
                author,
                "--author-timestamp",
                author_date,
            ]
        )

        check_call(["jj", "new", "--quiet"])

        if not self.args.no_branch and config.create_branch:
            # # Advance the bookmark we created for this patch. Because we know we're creating
            # entirely new commits, and we have the branch at the parent commit, we can let `--from`
            # compensate for the fact that we otherwise don't remember the name of the branch.

            check_call(
                [
                    "jj",
                    "bookmark",
                    "move",
                    "--quiet",
                    self.__patch_branch_name,
                    "--to=@-",
                ]
            )

    def format_patch(
        self, diff: str, body: str, author: Optional[str], author_date: Optional[int]
    ) -> str:
        return diff

    # ----
    # Methods private to this abstraction.
    # ----

    def __cli_log(
        self, *, revset: str, template: str, use_reversed: bool = False, **kwargs
    ) -> Union[List[str], str]:
        options = []
        if use_reversed:
            options.append("--reversed")
        return self.__check_output(
            [
                "jj",
                "log",
                "--no-pager",
                "--no-graph",
                "--revisions",
                revset,
                "--template",
                template,
            ]
            + options
            + ["--"],
            **kwargs,
        )

    def __validate_email(self):
        """Validate the user's configured email (user.email)."""

        if not is_valid_email(self.__email):
            raise Error(
                f"Your email configured with Jujutsu ({self.git.email}) is not a valid "
                f"format.\n"
                f"Please run `jj config set user.email someone@example.com …` to set "
                f"the correct value.\n"
                # NOTE: Unlike with a Git repository, it's not easy to specify a custom author from
                # the CLI. There is upstream work to track a better CLI for this:
                # <https://github.com/martinvonz/jj/issues/4170>
            )

    def __get_last_stack_change(self) -> Optional[str]:
        """Gets the last of the current stack of mutable changes, but _only_ if there's one."""
        # TODO: Should we do something different when `config.git_remote` or `self.args.upstream`
        # are specified? Compare with `remotes` checks in Git impl.

        revset = 'heads(immutable()..@- | present(@ ~ description(exact:"")))'
        mutable_roots = self.__cli_log(template='change_id ++ "\\n"', revset=revset)
        if not mutable_roots:
            return None
        elif len(mutable_roots) > 1:
            raise Error(
                f"Multiple mutable heads found (revset `{revset}`), unable to continue"
            )
        return mutable_roots[0]

    def __get_first_stack_change(self) -> Optional[str]:
        """Gets the first of the current stack of mutable changes, but _only_ if there's one."""
        # TODO: Should we do something different when `config.git_remote` or `self.args.upstream`
        # are specified? Compare with `remotes` checks in Git impl.

        revset = 'roots(immutable()..@- | present(@ ~ description(exact:"")))'
        mutable_roots = self.__cli_log(template='change_id ++ "\\n"', revset=revset)
        if not mutable_roots:
            return None
        elif len(mutable_roots) > 1:
            raise Error(
                f"Multiple mutable roots found (revset `{revset}`), unable to continue"
            )
        return mutable_roots[0]

    @staticmethod
    def __parse_log_bool(name: str, s: str) -> bool:
        if s not in ["true", "false"]:
            raise Error(f"internal error: {name} was not `true` or `false`")
        return s == "true"

    def __check_output(self, *args, **kwargs) -> Union[List[str], str]:
        return check_output(*args, stderr=subprocess.PIPE, **kwargs)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import annotations

import json
import os
import re
import stat
import sys
import tempfile
from contextlib import contextmanager
from itertools import zip_longest
from shutil import which
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
)

from mozphab import environment

from .commits import Commit
from .logger import logger
from .simplecache import cache

# If a commit body matches **all** of these, reject it.  This is to avoid the
# necessity to merge arc-style fields across an existing commit description
# and what we need to set.
ARC_REJECT_RE_LIST = [
    re.compile(r"^Summary:", flags=re.MULTILINE),
    re.compile(r"^Reviewers:", flags=re.MULTILINE),
]


# Regex for matching Phabricator URLs with revision
PHABRICATOR_URL_REVISION_PATTERN = r"https?://[^/]+/D(?P<rev>\d+)"

ARC_DIFF_REV_RE = re.compile(
    r"^\s*Differential Revision:\s*" + PHABRICATOR_URL_REVISION_PATTERN + r"\s*$",
    flags=re.MULTILINE,
)
ORIGINAL_DIFF_REV_RE = re.compile(
    r"^\s*Original Revision:\s*" + PHABRICATOR_URL_REVISION_PATTERN + r"\s*$",
    flags=re.MULTILINE,
)

# Bug and review regexs (from vct's commitparser)
BUG_ID_RE = re.compile(r"(?:(?:bug|b=)(?:\s*)(\d+)(?=\b))", flags=re.IGNORECASE)
LIST = r"[;,\/\\]\s*"
LIST_RE = re.compile(LIST)
IRC_NICK_CHARS = r"a-zA-Z0-9\-\_!"  # includes !, which is different from commitparser
IRC_NICK_CHARS_WITH_PERIOD = IRC_NICK_CHARS + r"."
IRC_NICK = r"#?[" + IRC_NICK_CHARS_WITH_PERIOD + r"]*[" + IRC_NICK_CHARS + r"]+"
REVIEWERS_RE = (
    r"([\s(.\[;,])(r%s)("
    + IRC_NICK
    + r"(?:"
    + LIST
    + r"(?![a-z0-9.\-]+[=?])"
    + IRC_NICK
    + r")*)?"
)
ALL_REVIEWERS_RE = re.compile(REVIEWERS_RE % r"[=?]")
REQUEST_REVIEWERS_RE = re.compile(REVIEWERS_RE % r"[?]")
GRANTED_REVIEWERS_RE = re.compile(REVIEWERS_RE % r"=")
R_SPECIFIER_RE = re.compile(r"\br[=?]")
BLOCKING_REVIEWERS_RE = re.compile(r"\b(r!)([" + IRC_NICK_CHARS_WITH_PERIOD + ",]+)")

DEPENDS_ON_RE = re.compile(r"^\s*Depends on\s*D(\d+)\s*$", flags=re.MULTILINE)

DONTBUILD_RE = re.compile(r"\s*\(?DONTBUILD\)?(?:\s*\(NPOTB\))?")


VALID_EMAIL_RE = re.compile(r"[^@ \t\r\n]+@[^@ \t\r\n]+\.[^@ \t\r\n]+")


def which_path(path: str) -> Optional[str]:
    """Check if an executable is provided. Fall back to which if not.

    Args:
        path: (str) filename or path to check for an executable command

    Returns:
        The full path of a command or None.
    """
    if (
        os.path.exists(path)
        and os.access(path, os.F_OK | os.X_OK)
        and not os.path.isdir(path)
    ):
        logger.debug("Path found: %s", path)
        return path

    return which(path)


def parse_config(
    config_list: List[str], filter_func: Optional[Callable] = None
) -> dict:
    """Parses list with "name=value" strings.

    Args:
        config_list: A list of "name=value" strings
        filter_func: A function taking the parsing config name and value for each line.
            If the function returns True the config value will be included in the final
            dict.

    Returns: A dict containing parsed data.
    """
    result = {}
    for line in config_list:
        try:
            name, value = line.split("=", 1)
        except ValueError:
            continue

        name = name.strip()
        value = value.strip()
        if filter_func is None or (callable(filter_func) and filter_func(name, value)):
            result[name] = value

    return result


def read_json_field(files: List[str], field_path: List[str]) -> Optional[str]:
    """Parses json files in turn returning value as per field_path, or None."""
    for filename in files:
        try:
            with open(filename, encoding="utf-8") as f:
                rc = json.load(f)
                for field_name in field_path:
                    if field_name not in rc:
                        rc = None
                        break
                    rc = rc[field_name]
                if not rc:
                    continue
                return rc
        except FileNotFoundError:
            continue
        except ValueError:
            continue
    return None


@contextmanager
def temporary_file(
    content: str, encoding: str = "utf-8", newline: Optional[str] = "\n"
):
    f = tempfile.NamedTemporaryFile(
        delete=False, mode="w+", encoding=encoding, newline=newline
    )
    try:
        f.write(content)
        f.flush()
        f.close()
        yield f.name
    finally:
        os.remove(f.name)


@contextmanager
def temporary_binary_file(content: bytes):
    f = tempfile.NamedTemporaryFile(delete=False, mode="wb+")
    try:
        f.write(content)
        f.flush()
        f.close()
        yield f.name
    finally:
        os.remove(f.name)


def get_arcrc_path() -> str:
    """Return a path to the user's Arcanist configuration file."""
    if "arcrc" in cache:
        return str(cache.get("arcrc"))

    if environment.IS_WINDOWS:
        arcrc = os.path.join(os.getenv("APPDATA", ""), ".arcrc")
    else:
        arcrc = os.path.expanduser("~/.arcrc")
        if os.path.isfile(arcrc) and stat.S_IMODE(os.stat(arcrc).st_mode) != 0o600:
            logger.debug("Changed file permissions on the %s file.", arcrc)
            os.chmod(arcrc, 0o600)

    cache.set("arcrc", arcrc)
    return arcrc


def parse_arc_diff_rev(body: str) -> Optional[int]:
    m = ARC_DIFF_REV_RE.search(body)
    return int(m.group("rev")) if m else None


def strip_differential_revision(body: str) -> str:
    return ARC_DIFF_REV_RE.sub("", body).rstrip()


def strip_dontbuild(text: str) -> str:
    """Remove the DONTBUILD flag from text."""
    return DONTBUILD_RE.sub("", text).strip()


def move_drev_to_original(
    body: str, rev_id: Optional[int]
) -> Tuple[str, Optional[int]]:
    """Handle moving the `Differential Revision` line.

    Moves the `Differential Revision` line to `Original Revision`, if a link
    to the original revision does not already exist. If the `Original Revision`
    line does exist, scrub the `Differential Revision` line.

    Args:
        body: `str` text of the commit message.
        rev_id: `int` parsed integer representing the drev number for the revision.

    Returns:
        tuple of:
            New commit message body text as `str`,
            Revision id as `int`, or `None` if a new revision should be created.
    """
    # Previous logic did not find a revision id, so this function won't either.
    if not rev_id:
        return body, rev_id

    original_revision = ORIGINAL_DIFF_REV_RE.search(body)
    if original_revision:
        return strip_differential_revision(body), None

    # Replace "Differential Revision" with "Original Revision", and set the
    # `rev-id` to `None`
    return (
        ARC_DIFF_REV_RE.sub(
            lambda m: m.group(0).replace(
                "Differential Revision:", "Original Revision:"
            ),
            body,
        ),
        None,
    )


def parse_api_error(api_response: str) -> Optional[str]:
    """Parse the string output from 'arc call-conduit' and return any errors found.

    Args:
        api_response: stdout string captured from 'arc call-conduit'.  It should
            contain only valid JSON.

    Returns:
        A string error description if an error occurred or None if no error occurred.
    """
    # Example error response from running
    # $ echo '{}' | arc call-conduit differential.revision.edit | jq .
    #
    # {
    #   "error": "ERR-CONDUIT-CORE",
    #   "errorMessage": "ERR-CONDUIT-CORE: Parameter \"transactions\" is not a list "\
    #       "of transactions.",
    #   "response": null
    # }
    response = json.loads(api_response)
    if response["error"] and response["errorMessage"]:
        return response["errorMessage"]


def prompt(question: str, options: Optional[List[str]] = None):
    if environment.HAS_ANSI:
        question = "\033[33m%s\033[0m" % question
    prompt_str = question
    if options:
        prompt_options = list(options)
        prompt_options[0] = prompt_options[0].upper()
        prompt_str = "%s (%s)? " % (question, "/".join(prompt_options))

        options_map = {o[0].lower(): o for o in options}
        options_map[""] = options[0]

    while True:
        res = input(prompt_str)

        if res == chr(27):  # escape
            sys.exit(1)

        if not options:
            return res

        if len(res) > 1:
            res = res[0].lower()

        if res in options_map:
            return options_map[res]


def has_arc_rejections(body: str) -> bool:
    return all(r.search(body) for r in ARC_REJECT_RE_LIST)


def augment_commits_from_body(commits: List[Commit]):
    """Extract metadata from commit body as fields.

    Adds: rev-id, bug-id, reviewers, wip
    """
    for commit in commits:
        commit.rev_id = parse_arc_diff_rev(commit.body)

        # bug-id
        bug_ids = parse_bugs(commit.title)
        if bug_ids:
            if len(bug_ids) > 1:
                logger.warning("Multiple bug-IDs found, using %s", bug_ids[0])
            commit.bug_id = bug_ids[0]
        else:
            commit.bug_id = None

        if not commit.bug_id_orig:
            commit.bug_id_orig = commit.bug_id

        # Set reviewers from title.
        commit.reviewers = parse_reviewers(commit.title)

        # Mark commit as WIP if commit desc starts with "WIP:".
        commit.wip = commit.wip_in_commit_title()


def parse_bugs(title: str) -> List[str]:
    return list(BUG_ID_RE.findall(title))


def parse_reviewers(title: str) -> dict:
    """Extract reviewers information from first line of the commit message.

    Returns a dictionary containing reviewers divided by the type:
        "r?" reviewers under the "request" key
        "r=" reviewers under the "granted" key
    """

    def extend_matches(match_re: re.Pattern, matches: List[str]):
        """Extends `matches` with any matches found using `match_re`.
        Args:
            match_re (re.Pattern): a regular expression object to search with
            matches (list of str): a list of strings of reviewers captured
        Returns:
            dict (str, list of str): a dictionary of requested and granted reviewers
        """
        for match in re.finditer(match_re, title):
            if match.group(3):
                matches.extend(re.split(LIST_RE, match.group(3)))

    reviewers = {"request": [], "granted": []}
    extend_matches(REQUEST_REVIEWERS_RE, reviewers["request"])
    extend_matches(GRANTED_REVIEWERS_RE, reviewers["granted"])
    return reviewers


def strip_depends_on(body: str) -> str:
    return DEPENDS_ON_RE.sub("", body).rstrip()


def update_commit_title_previews(commits: List[Commit]):
    """Update title-preview from commit metadata for all commits in stack"""
    for commit in commits:
        commit.title_preview = build_commit_title(commit)


def build_commit_title(commit: Commit) -> str:
    """Build/update title from commit metadata"""
    # Reviewers.
    title = replace_reviewers(commit.title, commit.reviewers)

    # Bug-ID.
    if commit.bug_id:
        if BUG_ID_RE.search(title):
            title = BUG_ID_RE.sub(f"Bug {commit.bug_id}", title, count=1)
        else:
            title = f"Bug {commit.bug_id} - {commit.title}"
    else:
        # This is likely to result in unappealing results.
        title = BUG_ID_RE.sub("", title)

    return title


def replace_reviewers(commit_description: str, reviewers: dict) -> str:
    """From vct's commitparser"""
    reviewers_lst = []
    if reviewers["request"]:
        reviewers_lst.append("r?" + ",".join(reviewers["request"]))

    if reviewers["granted"]:
        reviewers_lst.append("r=" + ",".join(reviewers["granted"]))

    reviewers_str = " ".join(reviewers_lst)

    if commit_description == "":
        return reviewers_str

    commit_description = commit_description.splitlines()
    commit_title = commit_description.pop(0)
    commit_description = "\n".join(commit_description)

    if not R_SPECIFIER_RE.search(commit_title):
        commit_title += " " + reviewers_str
    else:
        # replace the first r? with the reviewer list, and all subsequent
        # occurrences with a marker to mark the blocks we need to remove
        # later.
        d = {"first": True}

        def replace_first_reviewer(matchobj):
            if R_SPECIFIER_RE.match(matchobj.group(2)):
                if d["first"]:
                    d["first"] = False
                    return matchobj.group(1) + reviewers_str
                else:
                    return "\0"
            else:
                return matchobj.group(0)

        commit_title = re.sub(ALL_REVIEWERS_RE, replace_first_reviewer, commit_title)

        # remove marker values as well as leading separators.  this allows us
        # to remove runs of multiple reviewers and retain the trailing
        # separator.
        commit_title = re.sub(LIST + "\0", "", commit_title)
        commit_title = re.sub("\0", "", commit_title)

    if commit_description == "":
        return commit_title.strip()
    else:
        return commit_title.strip() + "\n" + commit_description


def prepare_body(
    title: str,
    summary: str,
    rev_id: int,
    phab_url: str,
) -> str:
    """Prepare the body using title and summary."""
    summary = strip_differential_revision(summary)
    summary = strip_depends_on(summary)

    if summary:
        summary += "\n\n"

    summary += f"Differential Revision: {phab_url}/D{rev_id}"
    body = f"{title}\n\n{summary}"

    return body


def short_node(text: str) -> str:
    """Return a shortened version of a node name."""
    if len(text) == 40:
        try:
            int(text, 16)
            return text[:12]
        except ValueError:
            pass
    return text


def create_hunk_lines(
    body: str, prefix: str, check_eof: bool = True
) -> Tuple[List[str], Optional[bool]]:
    """Parse a text body into a list of lines to be used in hunks.

    Args:
        body: A string containing the raw content of the file.
        prefix: A character (e.g. "+") indicating whether the lines are to be
            added or removed or are unchanged.
        check_eof: Whether to check or not to check the end of the file for the
            newline character. If set to True, the check is performed and if a
            newline character is missing, it will be added to the last entry in
            `lines` along with a message.

    Returns:
        A tuple containing a list of strings to be used when generating the corpus
            of the hunks, as well as boolean representing whether the file
            terminated with a new line or not if applicable, `None` if not.

    Note:
        When the body is missing a newline at the end, the appended lines will use a
        POSIX line terminator regardless of what line separators are already used in the
        body.
    """
    allowed_prefixes = ("+", "-", " ")
    if prefix not in allowed_prefixes:
        raise ValueError(f"Prefix should be one of {allowed_prefixes}")

    empty_file = body in ("", b"")

    if empty_file:
        # `body` has absolutely nothing in it, return values accordingly.
        if check_eof and prefix != "+":
            return ["\\ No newline at end of file\n"], True
        else:
            return [], None

    # Split lines on line separators
    lines = split_lines(body)
    eof_missing_newline = lines[-1] != ""

    # Re-add line separators for all lines except the last line.
    last_line = lines.pop()
    lines = join_lineseps(lines)
    lines = [f"{prefix}{line}" for line in lines]
    if eof_missing_newline:
        if check_eof:
            # Re-add the last line with the line separator, and another line
            # indicating that no new line was at the EOF.
            lines.append(f"{prefix}{last_line}\n")
            lines.append("\\ No newline at end of file\n")
        else:
            # Re-add the last line as is since we weren't supposed to touch it.
            lines.append(f"{prefix}{last_line}")

    return lines, eof_missing_newline if check_eof else None


def split_lines(body: bytes | str) -> List[bytes | str]:
    """Split a string on line separators, and keep line endings.

    This method behaves the same as `str.splitlines(True)`, but only splits on POSIX
    and DOS style line endings.

    Args:
        body: An arbitrary string.

    Returns:
        A list of strings consisting of lines and line separators.

    Example:
        >>> test = "line1\nline2\r\nline3"
        >>> split_lines(test)
        >>> ["line1", "\n", "line2", "\r\n", "line3"]
    """
    binary = isinstance(body, bytes)
    pattern = b"(\n|\r\n)"
    if not binary:
        pattern = pattern.decode("utf-8")

    return re.split(pattern, body)


def join_lineseps(lines: List[str]) -> List[str]:
    """Given a list of strings, join every two entries together where possible.

    NOTE: If the list length is odd, then the last entry is joined with an empty string.

    Args:
        lines: A list of strings representing lines to join.

    Returns:
        A list of strings that contains a joined list of strings.

    Example:
        >>> test = ["line1", "\n", "line2", "\r\n", "line3"]
        >>> join_lineseps(test)
        >>> ["line1\n", "line2\r\n", "line3"]
    """
    return [
        "".join(line) for line in zip_longest(lines[0::2], lines[1::2], fillvalue="")
    ]


def is_valid_email(email: str) -> bool:
    """Given a string, determines if it is a valid email."""
    return VALID_EMAIL_RE.match(email) is not None
